"""Module initialisation file

Code inside this file will be executed when this submodule is imported.

For contributors: feel free to add your initialisation code here if needed.
"""

from ._slug_test import *

__all__ = [ "SlugTest" ]
