c
c	Parameter file for RFI subroutines
c
        parameter       (maxlayer = 15,
c    &                   maxsublayer = 100,
     &                   maxsublayer = 25,
     &                   max_nd = maxlayer * 4,
     &                   maxmoddim = max_nd,
     &                   maxdata = 1024,
     &                   maxwave = 10 )

