/* https://github.com/pydata/pydata-sphinx-theme/blob/master/src/pydata_sphinx_theme/theme/pydata_sphinx_theme/layout.html */
/* https://github.com/executablebooks/sphinx-book-theme/blob/master/src/sphinx_book_theme/theme/sphinx_book_theme/layout.html */

/* ALL: color, text font size */
html[data-theme="light"] {
    --pst-color-primary: #E29578;
}
p {
    font-size: 1rem;
}


/* APV modifications */
/* Mainly designed to get FAQ looking good */
h1 {
    font-size: 200%;
}

h2 {
    font-size: 150%;
}

h3 {
    font-size:100%;
    font-weight: 450;
}

div.contents {
    padding: 1em 1em 1em;
}
div.contents>ul>li>p>a {
    font-size: 100%;
}
/* End APV modifications */


/* TOP LEFT: logo + project title */
div.navbar-brand-box a.navbar-brand img {
    max-height: 10vh !important;
}
span.project-version {
    color: var(--pst-color-primary);
    font-size: 0.8em;
    font-weight: 600;
    margin-left: 3px;
}

/* LEFT: side bar font size */
.bd-sidenav a.reference.internal {
    font-size: 0.8rem;
}

/* Mermaid */
div.mermaid:not([data-processed]) {
    display: none;
}
div.mermaid {
    transition: transform .2s;
}
div.mermaid:hover {
    transform: translateX(15%) scale(1.3);
}


/* HOME: panel button */
.card-border:hover {
    border: 1.5px solid var(--pst-color-primary) !important;
}

/* API: class/method name */
span.sig-name {
    color: #99582a;
}
code.xref {
    text-decoration: underline;
}
code.literal.docutils {
    color: #99582a;
}
table.autosummary p {
    font-size: 0.8rem;
}
table.autosummary span {
    font-size: 1rem;
}

/* TOC .. contents:: */
#developer-notes section div.contents {
    border: 0;
    background-color: transparent;
    padding: 0;
}
#developer-notes section div.contents ul {
    list-style-type: square;
}
#developer-notes div.contents > ul > li > p > a {
    font-size: 1em;
}
